//=================================================================================================
/*!
//  \file blaze/util/DimensionOf.h
//  \brief Compile time evaluation of array sizes
//
//  Copyright (C) 2012-2020 Klaus Iglberger - All Rights Reserved
//
//  This file is part of the Blaze library. You can redistribute it and/or modify it under
//  the terms of the New (Revised) BSD License. Redistribution and use in source and binary
//  forms, with or without modification, are permitted provided that the following conditions
//  are met:
//
//  1. Redistributions of source code must retain the above copyright notice, this list of
//     conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice, this list
//     of conditions and the following disclaimer in the documentation and/or other materials
//     provided with the distribution.
//  3. Neither the names of the Blaze development group nor the names of its contributors
//     may be used to endorse or promote products derived from this software without specific
//     prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
//  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
//  SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
//  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
//  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
//  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
//  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
//  DAMAGE.
*/
//=================================================================================================

#ifndef _BLAZE_UTIL_DIMENSIONOF_H_
#define _BLAZE_UTIL_DIMENSIONOF_H_


//*************************************************************************************************
// Includes
//*************************************************************************************************

#include <blaze/util/MaybeUnused.h>
#include <blaze/util/Types.h>


namespace blaze {

//=================================================================================================
//
//  DIMENSIONOF FUNCTIONALITY
//
//=================================================================================================

//*************************************************************************************************
/*!\brief Static evaluation of array dimensions.
// \ingroup util
//
// \param a Reference to a static array of type T and size N.
// \return Dimension of the static array.
//
// The dimensionof function is a safe way to evaluate the size of an array. The function only
// works for array arguments and fails for pointers and user-defined class types.

   \code
   int              ai[ 42 ];
   int*             pi( ai );
   std::vector<int> vi( 42 );

   dimensionof( ai );  // Returns the size of the integer array (42)
   dimensionof( pi );  // Fails to compile!
   dimensionof( vi );  // Fails to compile!
   \endcode
*/
template< typename T, size_t N >
constexpr size_t dimensionof( T(&a)[N] )
{
   MAYBE_UNUSED( a );
   return N;
}
//*************************************************************************************************

} // namespace blaze

#endif
