//=================================================================================================
/*!
//  \file blaze/math/typetraits/IsStatic.h
//  \brief Header file for the IsStatic type trait
//
//  Copyright (C) 2012-2020 Klaus Iglberger - All Rights Reserved
//
//  This file is part of the Blaze library. You can redistribute it and/or modify it under
//  the terms of the New (Revised) BSD License. Redistribution and use in source and binary
//  forms, with or without modification, are permitted provided that the following conditions
//  are met:
//
//  1. Redistributions of source code must retain the above copyright notice, this list of
//     conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice, this list
//     of conditions and the following disclaimer in the documentation and/or other materials
//     provided with the distribution.
//  3. Neither the names of the Blaze development group nor the names of its contributors
//     may be used to endorse or promote products derived from this software without specific
//     prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
//  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
//  SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
//  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
//  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
//  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
//  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
//  DAMAGE.
*/
//=================================================================================================

#ifndef _BLAZE_MATH_TYPETRAITS_ISSTATIC_H_
#define _BLAZE_MATH_TYPETRAITS_ISSTATIC_H_


//*************************************************************************************************
// Includes
//*************************************************************************************************

#include <blaze/math/typetraits/IsMatrix.h>
#include <blaze/math/typetraits/IsVector.h>
#include <blaze/math/typetraits/Size.h>
#include <blaze/util/IntegralConstant.h>


namespace blaze {

//=================================================================================================
//
//  CLASS DEFINITION
//
//=================================================================================================

//*************************************************************************************************
/*!\brief Compile time check for static data types.
// \ingroup math_type_traits
//
// This type trait tests whether the given data type is a static data type, i.e. a vector or
// matrix with dimensions fixed at compile time. In case the data type a static data type, the
// \a value member constant is set to \a true, the nested type definition \a Type is \a TrueType,
// and the class derives from \a TrueType. Otherwise \a value is set to \a false, \a Type is
// \a FalseType, and the class derives from \a FalseType. Examples:

   \code
   using blaze::StaticVector;
   using blaze::StaticMatrix;
   using blaze::DynamicVector;
   using blaze::DynamicMatrix;
   using blaze::columnVector;
   using blaze::rowMajor;

   using StaticVectorType = StaticVector<int,columnVector>;
   using StaticMatrixType = StaticMatrix<double,rowMajor>;

   blaze::IsStatic< StaticVectorType >::value                        // Evaluates to 1
   blaze::IsStatic< const StaticVectorType >::Type                   // Results in TrueType
   blaze::IsStatic< volatile StaticMatrixType >                      // Is derived from TrueType
   blaze::IsStatic< int >::value                                     // Evaluates to 0
   blaze::IsStatic< const DynamicVector<float,columnVector> >::Type  // Results in FalseType
   blaze::IsStatic< volatile DynamicMatrix<int,rowMajor> >           // Is derived from FalseType
   \endcode
*/
template< typename T >
struct IsStatic
   : public BoolConstant< ( IsVector_v<T> && Size_v<T,0UL> != DefaultSize_v ) ||
                          ( IsMatrix_v<T> && Size_v<T,0UL> != DefaultSize_v &&
                                             Size_v<T,1UL> != DefaultSize_v ) >
{};
//*************************************************************************************************


//*************************************************************************************************
/*!\brief Auxiliary variable template for the IsStatic type trait.
// \ingroup math_type_traits
//
// The IsStatic_v variable template provides a convenient shortcut to access the nested \a value
// of the IsStatic class template. For instance, given the type \a T the following two statements
// are identical:

   \code
   constexpr bool value1 = blaze::IsStatic<T>::value;
   constexpr bool value2 = blaze::IsStatic_v<T>;
   \endcode
*/
template< typename T >
constexpr bool IsStatic_v = IsStatic<T>::value;
//*************************************************************************************************

} // namespace blaze

#endif
