//=================================================================================================
/*!
//  \file blaze/util/TypeList.h
//  \brief Header file for the Unique class template
//
//  Copyright (C) 2012-2020 Klaus Iglberger - All Rights Reserved
//
//  This file is part of the Blaze library. You can redistribute it and/or modify it under
//  the terms of the New (Revised) BSD License. Redistribution and use in source and binary
//  forms, with or without modification, are permitted provided that the following conditions
//  are met:
//
//  1. Redistributions of source code must retain the above copyright notice, this list of
//     conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice, this list
//     of conditions and the following disclaimer in the documentation and/or other materials
//     provided with the distribution.
//  3. Neither the names of the Blaze development group nor the names of its contributors
//     may be used to endorse or promote products derived from this software without specific
//     prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
//  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
//  SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
//  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
//  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
//  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
//  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
//  DAMAGE.
*/
//=================================================================================================

#ifndef _BLAZE_UTIL_TYPELIST_UNIQUE_H_
#define _BLAZE_UTIL_TYPELIST_UNIQUE_H_


//*************************************************************************************************
// Includes
//*************************************************************************************************

#include <blaze/util/typelist/Append.h>
#include <blaze/util/typelist/Erase.h>
#include <blaze/util/typelist/TypeList.h>


namespace blaze {

//=================================================================================================
//
//  CLASS DEFINITION
//
//=================================================================================================

//*************************************************************************************************
/*!\brief Erasing all duplicates from a type list.
// \ingroup typelist
//
// The Unique class can be used to erase all duplicates from a type list \a TList. In order to
// erase all duplicates, the Unique class has to be instantiated for a particular type list.
// The following example gives an impression of the use of the Unique class:

   \code
   // Defining a temporary type list containing the types int and float twice
   using Tmp = blaze::TypeList< float, int, double, int, float >;

   // Removing all duplicates from the type list
   using NoDuplicates = blaze::Unique<Tmp>::Type;
   \endcode
*/
template< typename TL >  // Type of the type list
struct Unique;
//*************************************************************************************************


//*************************************************************************************************
/*! \cond BLAZE_INTERNAL */
/*!\brief Spezialization of the Unique class for an empty type list.
// \ingroup typelist
*/
template<>
struct Unique< TypeList<> >
{
   using Type = TypeList<>;
};
/*! \endcond */
//*************************************************************************************************


//*************************************************************************************************
/*! \cond BLAZE_INTERNAL */
/*!\brief Spezialization of the Unique class for a general type list.
// \ingroup typelist
*/
template< typename T        // Type of the head of the type list
        , typename... Ts >  // Types of the tail of the type list
struct Unique< TypeList<T,Ts...> >
{
 private:
   using TL1  = typename Unique< TypeList<Ts...> >::Type;
   using TL2  = typename Erase<TL1,T>::Type;

 public:
   using Type = typename Append< TypeList<T>, TL2 >::Type;
};
/*! \endcond */
//*************************************************************************************************


//*************************************************************************************************
/*!\brief Auxiliary alias declaration for the Unique class template.
// \ingroup type_traits
//
// The Unique_t alias declaration provides a convenient shortcut to access the nested \a Type
// of the Unique class template. For instance, given the type list \a TL the following two type
// definitions are identical:

   \code
   using Type1 = typename Unique<TL>::Type;
   using Type2 = Unique_t<TL>;
   \endcode
*/
template< typename TL >  // Type of the type list
using Unique_t = typename Unique<TL>::Type;
//*************************************************************************************************

} // namespace blaze

#endif
